@@ -14,6 +14,8 @@ from mcdreforged.command.command_manager import CommandManager
 from mcdreforged.config import Config
 from mcdreforged.executor.console_handler import ConsoleHandler
 from mcdreforged.executor.task_executor import TaskExecutor
+from mcdreforged.executor.update_helper import UpdateHelper
+from mcdreforged.executor.watchdog import WatchDog
 from mcdreforged.handler.server_handler_manager import ServerHandlerManager
 from mcdreforged.info import Info
 from mcdreforged.info_reactor.info_reactor_manager import InfoReactorManager
@@ -27,24 +29,23 @@ from mcdreforged.plugin.server_interface import ServerInterface
 from mcdreforged.utils import logger, file_util
 from mcdreforged.utils.exception import IllegalCallError, ServerStopped, ServerStartError, IllegalStateError
 from mcdreforged.utils.logger import DebugOption, MCDReforgedLogger
-from mcdreforged.utils.update_helper import UpdateHelper
 
 
 class MCDReforgedServer:
 	process: Optional[Popen]
 
-	def __init__(self, old_process=None):
+	def __init__(self):
 		# TODO console args
 		self.mcdr_state = MCDReforgedState.INITIALIZING
 		self.server_state = ServerState.STOPPED
-		self.process = old_process  # the process for the server
+		self.process = None  # type: Optional[PIPE]
 		self.flags = MCDReforgedFlag.NONE
 		self.starting_server_lock = Lock()  # to prevent multiple start_server() call
 		self.stop_lock = Lock()  # to prevent multiple stop() call
 
-		# will be assigned in reload_config()
-		self.encoding_method = None
-		self.decoding_method = None
+		# will be assigned in on_config_changed()
+		self.encoding_method = None  # type: Optional[str]
+		self.decoding_method = None  # type: Optional[str]
 
 		# --- Constructing fields ---
 		self.logger = MCDReforgedLogger(self)
@@ -52,6 +53,8 @@ class MCDReforgedServer:
 		self.server_interface = ServerInterface(self)
 		self.task_executor = TaskExecutor(self)
 		self.console_handler = ConsoleHandler(self)
+		self.watch_dog = WatchDog(self)
+		self.update_helper = UpdateHelper(self)
 		self.language_manager = LanguageManager(self.logger)
 		self.config = Config(self.logger)
 		self.rcon_manager = RconManager(self)
@@ -62,7 +65,6 @@ class MCDReforgedServer:
 		self.permission_manager = PermissionManager(self)
 
 		# --- Initialize fields instance ---
-		# register handlers first, so when the config is loaded the handler is ready to be set
 		file_missing = False
 		try:
 			self.load_config(allowed_missing_file=False)  # loads config, language, handlers
@@ -81,9 +83,6 @@ class MCDReforgedServer:
 			return
 		self.plugin_manager.register_permanent_plugins()
 
-		self.update_helper = UpdateHelper(self)
-		self.update_helper.check_update_start()
-
 		self.set_mcdr_state(MCDReforgedState.INITIALIZED)
 
 	def __del__(self):
@@ -149,7 +148,7 @@ class MCDReforgedServer:
 
 	def load_plugins(self):
 		self.plugin_manager.refresh_all_plugins()
-		self.logger.info(self.plugin_manager.last_operation_result.to_rtext())
+		self.logger.info(self.plugin_manager.last_operation_result.to_rtext(show_path=True))
 
 	def on_plugin_changed(self):
 		self.command_manager.clear_command()
@@ -160,7 +159,7 @@ class MCDReforgedServer:
 	# ---------------------------
 
 	def is_server_running(self):
-		return self.process is not None
+		return self.server_state.in_state({ServerState.RUNNING, ServerState.STOPPING})
 
 	# Flags
 
@@ -176,6 +175,9 @@ class MCDReforgedServer:
 	def is_mcdr_exit(self):
 		return self.mcdr_in_state(MCDReforgedState.STOPPED)
 
+	def is_mcdr_about_to_exit(self):
+		return self.mcdr_in_state({MCDReforgedState.PRE_STOPPED, MCDReforgedState.STOPPED})
+
 	def is_exit_naturally(self):
 		return MCDReforgedFlag.EXIT_NATURALLY in self.flags
 
@@ -230,21 +232,21 @@ class MCDReforgedServer:
 		"""
 		try to start the server process
 		return True if the server process has started successfully
-		return False if the server is already running or start_server has been called by other
+		return False if the server is not able to start
 
 		:return: a bool as above
 		:rtype: bool
 		"""
-		acquired = self.starting_server_lock.acquire(blocking=False)
-		if not acquired:
-			return False
-		try:
+		with self.starting_server_lock:
 			if self.is_interrupt():
 				self.logger.warning(self.tr('mcdr_server.start_server.already_interrupted'))
 				return False
 			if self.is_server_running():
 				self.logger.warning(self.tr('mcdr_server.start_server.start_twice'))
 				return False
+			if self.is_mcdr_about_to_exit():
+				self.logger.warning(self.tr('mcdr_server.start_server.about_to_exit'))
+				return False
 			cwd = self.config['working_directory']
 			if not os.path.isdir(cwd):
 				self.logger.error(self.tr('mcdr_server.start_server.cwd_not_existed', cwd))
@@ -259,8 +261,6 @@ class MCDReforgedServer:
 			else:
 				self.on_server_start()
 				return True
-		finally:
-			self.starting_server_lock.release()
 
 	def kill_server(self):
 		"""
@@ -321,10 +321,10 @@ class MCDReforgedServer:
 	# --------------------------
 
 	def on_server_start(self):
-		self.set_exit_naturally(True)
+		self.set_server_state(ServerState.RUNNING)
+		self.set_exit_naturally(True)  # Set after server state is set to RUNNING, or MCDR might have a chance to exit if the server is started by other thread
 		self.logger.info(self.tr('mcdr_server.start_server.pid_info', self.process.pid))
 		self.plugin_manager.dispatch_event(MCDRPluginEvents.SERVER_START, ())
-		self.set_server_state(ServerState.RUNNING)
 
 	def on_server_stop(self):
 		return_code = self.process.poll()
@@ -332,9 +332,9 @@ class MCDReforgedServer:
 		self.process.stdin.close()
 		self.process.stdout.close()
 		self.process = None
-		self.remove_flag(MCDReforgedFlag.SERVER_STARTUP | MCDReforgedFlag.SERVER_RCON_READY)  # removes this two
-		self.plugin_manager.dispatch_event(MCDRPluginEvents.SERVER_STOP, (return_code,))
 		self.set_server_state(ServerState.STOPPED)
+		self.remove_flag(MCDReforgedFlag.SERVER_STARTUP | MCDReforgedFlag.SERVER_RCON_READY)  # removes this two
+		self.plugin_manager.dispatch_event(MCDRPluginEvents.SERVER_STOP, (return_code,), wait=True)
 
 	def send(self, text, ending='\n', encoding=None):
 		"""
@@ -402,21 +402,21 @@ class MCDReforgedServer:
 			parsed_result = self.server_handler_manager.get_current_handler().parse_server_stdout(text)
 		except:
 			if self.logger.should_log_debug(option=DebugOption.HANDLER):  # traceback.format_exc() is costly
-				self.logger.debug('Fail to parse text "{}" from stdout of the server, using raw handler'.format(text))
+				self.logger.debug('Fail to parse text "{}" from stdout of the server, using raw handler'.format(text), no_check=True)
 				for line in traceback.format_exc().splitlines():
-					self.logger.debug('    {}'.format(line))
+					self.logger.debug('    {}'.format(line), no_check=True)
 			parsed_result = self.server_handler_manager.get_basic_handler().parse_server_stdout(text)
 		else:
 			if self.logger.should_log_debug(option=DebugOption.HANDLER):
-				self.logger.debug('Parsed text from server stdin:')
+				self.logger.debug('Parsed text from server stdout:', no_check=True)
 				for line in parsed_result.format_text().splitlines():
-					self.logger.debug('    {}'.format(line))
+					self.logger.debug('    {}'.format(line), no_check=True)
 		self.server_handler_manager.detect_text(text)
 		self.reactor_manager.put_info(parsed_result)
 
 	def on_mcdr_start(self):
 		self.task_executor.start()
-		self.task_executor.add_task(self.load_plugins)
+		self.task_executor.enqueue_regular_task(self.load_plugins)
 		self.task_executor.wait_till_finish_all_task()
 		self.plugin_manager.dispatch_event(MCDRPluginEvents.MCDR_START, ())
 		if not self.config['disable_console_thread']:
@@ -425,19 +425,22 @@ class MCDReforgedServer:
 			self.logger.info(self.tr('mcdr_server.on_mcdr_start.console_disabled'))
 		if not self.start_server():
 			raise ServerStartError()
+		self.update_helper.start()
+		self.watch_dog.start()
 		self.server_handler_manager.start_handler_detection()
 		self.set_mcdr_state(MCDReforgedState.RUNNING)
 
 	def on_mcdr_stop(self):
 		try:
+			self.set_mcdr_state(MCDReforgedState.PRE_STOPPED)
 			if self.is_interrupt():
 				self.logger.info(self.tr('mcdr_server.on_mcdr_stop.user_interrupted'))
 			else:
 				self.logger.info(self.tr('mcdr_server.on_mcdr_stop.server_stop'))
-			self.plugin_manager.dispatch_event(MCDRPluginEvents.MCDR_STOP, ())
 
-			self.set_mcdr_state(MCDReforgedState.PRE_STOPPED)
-			self.task_executor.join()
+			self.watch_dog.stop()  # it's ok for plugins to take some time
+			self.watch_dog.join()
+			self.plugin_manager.dispatch_event(MCDRPluginEvents.MCDR_STOP, (), wait=True)
 
 			self.logger.info(self.tr('mcdr_server.on_mcdr_stop.bye'))
 		except KeyboardInterrupt:  # I don't know why there sometimes will be a KeyboardInterrupt if MCDR is stopped by ctrl-c
