@@ -1,10 +1,12 @@
 """
 Plugin management
 """
+import collections
 import os
 import sys
 import threading
-from typing import Callable, Dict, Optional, Any, Tuple, List, TYPE_CHECKING
+from contextlib import contextmanager
+from typing import Callable, Dict, Optional, Any, Tuple, List, TYPE_CHECKING, Deque
 
 from mcdreforged import constant
 from mcdreforged.plugin.meta.dependency_walker import DependencyWalker
@@ -12,11 +14,12 @@ from mcdreforged.plugin.operation_result import PluginOperationResult, SingleOpe
 from mcdreforged.plugin.permanent.mcdreforged_plugin import MCDReforgedPlugin
 from mcdreforged.plugin.plugin import PluginState, AbstractPlugin
 from mcdreforged.plugin.plugin_event import MCDRPluginEvents, MCDREvent, EventListener
-from mcdreforged.plugin.plugin_registry import PluginManagerRegistry
+from mcdreforged.plugin.plugin_registry import PluginRegistryStorage
 from mcdreforged.plugin.plugin_thread import PluginThreadPool
 from mcdreforged.plugin.regular_plugin import RegularPlugin
 from mcdreforged.utils import file_util, string_util, misc_util
 from mcdreforged.utils.logger import DebugOption
+from mcdreforged.utils.thread_local_storage import ThreadLocalStorage
 
 if TYPE_CHECKING:
 	from mcdreforged.mcdr_server import MCDReforgedServer
@@ -25,6 +28,8 @@ PLUGIN_CONFIG_DIRECTORY = 'config'
 
 
 class PluginManager:
+	TLS_PLUGIN_KEY = 'current_plugin'
+
 	def __init__(self, mcdr_server: 'MCDReforgedServer'):
 		self.plugin_directories = []  # type: List[str]
 		self.mcdr_server = mcdr_server
@@ -35,7 +40,7 @@ class PluginManager:
 		# file_path -> id mapping
 		self.plugin_file_path = {}  # type: Dict[str, str]
 		# storage for event listeners, help messages and commands
-		self.registry_storage = PluginManagerRegistry(self)
+		self.registry_storage = PluginRegistryStorage(self)
 
 		self.last_operation_result = PluginOperationResult(self)
 
@@ -43,8 +48,10 @@ class PluginManager:
 		self.thread_pool = PluginThreadPool(self.mcdr_server, max_thread=constant.PLUGIN_THREAD_POOL_SIZE)
 
 		# thread local storage, to store current plugin
-		self.tls = threading.local()
-		self.set_current_plugin(None)
+		self.tls = ThreadLocalStorage()
+
+		# plugin manipulation lock
+		self.mani_lock = threading.RLock()
 
 		file_util.touch_directory(PLUGIN_CONFIG_DIRECTORY)
 
@@ -52,11 +59,13 @@ class PluginManager:
 	#   Getters / Setters etc.
 	# --------------------------
 
-	def get_current_running_plugin(self) -> Optional[RegularPlugin]:
+	def get_current_running_plugin(self, *, thread=None) -> Optional[AbstractPlugin]:
 		"""
 		Get current executing plugin in this thread
+		:param thread: If specified, it should be a Thread instance. Then it will return the executing plugin in the given thread
 		"""
-		return self.tls.current_plugin
+		stack = self.tls.get(self.TLS_PLUGIN_KEY, None, thread=thread)  # type: Deque[AbstractPlugin]
+		return stack[-1] if stack is not None else None
 
 	def get_all_plugins(self) -> List[AbstractPlugin]:
 		return list(self.plugins.values())
@@ -73,9 +82,6 @@ class PluginManager:
 			plugin = None
 		return plugin
 
-	def set_current_plugin(self, plugin: Optional[AbstractPlugin]):
-		self.tls.current_plugin = plugin
-
 	def set_plugin_directories(self, plugin_directories: Optional[List[str]]):
 		if plugin_directories is None:
 			plugin_directories = []
@@ -89,6 +95,18 @@ class PluginManager:
 			file_util.touch_directory(plugin_directory)
 			sys.path.append(plugin_directory)
 
+	@contextmanager
+	def with_plugin_context(self, plugin: AbstractPlugin):
+		stack = self.tls.get(self.TLS_PLUGIN_KEY, default=collections.deque())  # type: Deque[AbstractPlugin]
+		stack.append(plugin)
+		self.tls.put(self.TLS_PLUGIN_KEY, stack)
+		try:
+			yield
+		finally:
+			stack.pop()
+			if len(stack) == 0:
+				self.tls.pop(self.TLS_PLUGIN_KEY)
+
 	def contains_plugin_file(self, file_path: str) -> bool:
 		return file_path in self.plugin_file_path
 
@@ -116,15 +134,20 @@ class PluginManager:
 
 	def __add_plugin(self, plugin: AbstractPlugin):
 		plugin_id = plugin.get_id()
+		if plugin_id in self.plugins:
+			self.logger.critical('Something is not correct, a plugin with existed plugin id "{}" is added'.format(plugin_id))
 		self.plugins[plugin_id] = plugin
 		if isinstance(plugin, RegularPlugin):
 			self.plugin_file_path[plugin.file_path] = plugin_id
 
 	def __remove_plugin(self, plugin: AbstractPlugin):
 		if not plugin.is_permanent():
-			self.plugins.pop(plugin.get_id())
+			plugin_id = plugin.get_id()
+			if plugin_id in self.plugins:
+				self.plugins.pop(plugin_id)
 			if isinstance(plugin, RegularPlugin):
-				self.plugin_file_path.pop(plugin.file_path)
+				if plugin.file_path in self.plugin_file_path:
+					self.plugin_file_path.pop(plugin.file_path)
 
 	# ----------------------------
 	#   Single Plugin Operations
@@ -148,8 +171,8 @@ class PluginManager:
 		else:
 			existed_plugin = self.plugins.get(plugin.get_id())
 			if existed_plugin is None:
-				self.logger.info(self.mcdr_server.tr('plugin_manager.load_plugin.success', plugin.get_name()))
 				self.__add_plugin(plugin)
+				self.logger.info(self.mcdr_server.tr('plugin_manager.load_plugin.success', plugin.get_name()))
 				return plugin
 			else:
 				self.logger.error(self.mcdr_server.tr('plugin_manager.load_plugin.duplicate', plugin.get_name(), plugin.file_path, existed_plugin.get_name(), existed_plugin.file_path))
@@ -192,22 +215,35 @@ class PluginManager:
 		:return: If the plugin reloads successfully without error
 		:rtype: bool
 		"""
+		plugin.receive_event(MCDRPluginEvents.PLUGIN_UNLOADED, ())
+		self.__remove_plugin(plugin)
 		try:
-			plugin.receive_event(MCDRPluginEvents.PLUGIN_UNLOADED, ())
 			plugin.reload()
 		except:
 			self.logger.exception(self.mcdr_server.tr('plugin_manager.reload_plugin.fail', plugin.get_name()))
 			self.__unload_plugin(plugin)
 			return False
 		else:
-			self.logger.info(self.mcdr_server.tr('plugin_manager.reload_plugin.success', plugin.get_name()))
-			return True
+			# in case the plugin id changes into an existed plugin id
+			existed_plugin = self.plugins.get(plugin.get_id())
+			if existed_plugin is None:
+				self.__add_plugin(plugin)
+				self.logger.info(self.mcdr_server.tr('plugin_manager.reload_plugin.success', plugin.get_name()))
+				return True
+			else:
+				self.logger.error(self.mcdr_server.tr('plugin_manager.load_plugin.duplicate', plugin.get_name(), plugin.file_path, existed_plugin.get_name(), existed_plugin.file_path))
+				try:
+					plugin.unload()
+				except:
+					# should never come here
+					self.logger.exception(self.mcdr_server.tr('plugin_manager.load_plugin.unload_duplication_fail', plugin.get_name(), plugin.file_path))
+				return False
 
 	# ---------------------------------------
 	#   Regular Plugin Collector & Handlers
 	# ---------------------------------------
 
-	def collect_and_process_new_plugins(self, filter: Callable[[str], bool]) -> SingleOperationResult:
+	def __collect_and_process_new_plugins(self, filter: Callable[[str], bool]) -> SingleOperationResult:
 		result = SingleOperationResult()
 		for plugin_directory in self.plugin_directories:
 			file_list = file_util.list_file_with_suffix(plugin_directory, constant.PLUGIN_FILE_SUFFIX)
@@ -220,7 +256,7 @@ class PluginManager:
 						result.succeed(plugin)
 		return result
 
-	def collect_and_remove_plugins(self, filter: Callable[[RegularPlugin], bool], specific: Optional[RegularPlugin] = None) -> SingleOperationResult:
+	def __collect_and_remove_plugins(self, filter: Callable[[RegularPlugin], bool], specific: Optional[RegularPlugin] = None) -> SingleOperationResult:
 		result = SingleOperationResult()
 		plugin_list = self.get_regular_plugins() if specific is None else [specific]
 		for plugin in plugin_list:
@@ -228,7 +264,7 @@ class PluginManager:
 				result.record(plugin, self.__unload_plugin(plugin))
 		return result
 
-	def reload_ready_plugins(self, filter: Callable[[RegularPlugin], bool], specific: Optional[RegularPlugin] = None) -> SingleOperationResult:
+	def __reload_ready_plugins(self, filter: Callable[[RegularPlugin], bool], specific: Optional[RegularPlugin] = None) -> SingleOperationResult:
 		result = SingleOperationResult()
 		plugin_list = self.get_regular_plugins() if specific is None else [specific]
 		for plugin in plugin_list:
@@ -236,7 +272,7 @@ class PluginManager:
 				result.record(plugin, self.__reload_plugin(plugin))
 		return result
 
-	def check_plugin_dependencies(self) -> SingleOperationResult:
+	def __check_plugin_dependencies(self) -> SingleOperationResult:
 		result = SingleOperationResult()
 		walker = DependencyWalker(self)
 		walk_result = walker.walk()
@@ -265,8 +301,8 @@ class PluginManager:
 	# Current behavior for plugin operations
 	# 1. Actual plugin operations
 	#   For plugin refresh
-	#     1. Load new plugins
-	#     2. Unload plugins whose file is removed
+	#     1. Unload plugins whose file is removed
+	#     2. Load new plugins
 	#     3. Reload existed (and matches filter) plugins whose state is ready. if reload fail unload it
 	#
 	#   For single plugin operation
@@ -277,10 +313,6 @@ class PluginManager:
 	#     2. Call on_load for new / reloaded plugins by topo order
 	#     3. Call on_unload for unloaded plugins
 
-	def __check_thread(self):
-		if not self.mcdr_server.task_executor.is_on_thread():
-			raise RuntimeError('Plugin operations should be executed directly on task executor\'s thread, but thread {} founded'.format(threading.current_thread().getName()))
-
 	def __post_plugin_process(self, load_result=None, unload_result=None, reload_result=None):
 		if load_result is None:
 			load_result = SingleOperationResult()
@@ -289,7 +321,7 @@ class PluginManager:
 		if reload_result is None:
 			reload_result = SingleOperationResult()
 
-		dependency_check_result = self.check_plugin_dependencies()
+		dependency_check_result = self.__check_plugin_dependencies()
 		self.last_operation_result.record(load_result, unload_result, reload_result, dependency_check_result)
 
 		# Expected plugin states:
@@ -299,6 +331,8 @@ class PluginManager:
 		# reload_result		READY				UNLOADING
 		# dep_chk_result	LOADED / READY		UNLOADING
 
+		self.registry_storage.clear()  # in case plugin invokes dispatch_event during on_load. dont let them trigger listeners
+
 		for plugin in load_result.success_list + reload_result.success_list:
 			if plugin in dependency_check_result.success_list:
 				plugin.ready()
@@ -321,14 +355,11 @@ class PluginManager:
 		for plugin in self.get_regular_plugins():
 			plugin.assert_state({PluginState.READY})
 
+		self.__sort_plugins_by_id()
 		self.__update_registry()
 
-	def __refresh_plugins(self, reload_filter: Callable[[RegularPlugin], bool]):
-		self.__check_thread()
-		load_result = self.collect_and_process_new_plugins(lambda fp: True)
-		unload_result = self.collect_and_remove_plugins(lambda plugin: not plugin.file_exists())
-		reload_result = self.reload_ready_plugins(reload_filter)
-		self.__post_plugin_process(load_result, unload_result, reload_result)
+	def __sort_plugins_by_id(self):
+		self.plugins = dict(sorted(self.plugins.items(), key=lambda item: item[0]))
 
 	def __update_registry(self):
 		self.registry_storage.clear()
@@ -337,68 +368,82 @@ class PluginManager:
 		self.registry_storage.arrange()
 		self.mcdr_server.on_plugin_changed()
 
+	def __refresh_plugins(self, reload_filter: Callable[[RegularPlugin], bool]):
+		unload_result = self.__collect_and_remove_plugins(lambda plugin: not plugin.file_exists())
+		load_result = self.__collect_and_process_new_plugins(lambda fp: True)
+		reload_result = self.__reload_ready_plugins(reload_filter)
+		self.__post_plugin_process(load_result, unload_result, reload_result)
+
 	# --------------
 	#   Interfaces
 	# --------------
 
 	def load_plugin(self, file_path: str):
-		self.__check_thread()
-		self.logger.info(self.mcdr_server.tr('plugin_manager.load_plugin.entered', file_path))
-		load_result = self.collect_and_process_new_plugins(lambda fp: fp == file_path)
-		self.__post_plugin_process(load_result=load_result)
+		with self.mani_lock:
+			self.logger.info(self.mcdr_server.tr('plugin_manager.load_plugin.entered', file_path))
+			load_result = self.__collect_and_process_new_plugins(lambda fp: fp == file_path)
+			self.__post_plugin_process(load_result=load_result)
 
 	def unload_plugin(self, plugin: RegularPlugin):
-		self.__check_thread()
-		self.logger.info(self.mcdr_server.tr('plugin_manager.unload_plugin.entered', plugin))
-		unload_result = self.collect_and_remove_plugins(lambda plg: True, specific=plugin)
-		self.__post_plugin_process(unload_result=unload_result)
+		with self.mani_lock:
+			self.logger.info(self.mcdr_server.tr('plugin_manager.unload_plugin.entered', plugin))
+			unload_result = self.__collect_and_remove_plugins(lambda plg: True, specific=plugin)
+			self.__post_plugin_process(unload_result=unload_result)
 
 	def reload_plugin(self, plugin: RegularPlugin):
-		self.__check_thread()
-		self.logger.info(self.mcdr_server.tr('plugin_manager.reload_plugin.entered', plugin))
-		reload_result = self.reload_ready_plugins(lambda plg: True, specific=plugin)
-		self.__post_plugin_process(reload_result=reload_result)
+		with self.mani_lock:
+			self.logger.info(self.mcdr_server.tr('plugin_manager.reload_plugin.entered', plugin))
+			reload_result = self.__reload_ready_plugins(lambda plg: True, specific=plugin)
+			self.__post_plugin_process(reload_result=reload_result)
 
 	def enable_plugin(self, file_path: str):
-		self.logger.info(self.mcdr_server.tr('plugin_manager.enable_plugin.entered', file_path))
-		new_file_path = string_util.remove_suffix(file_path, constant.DISABLED_PLUGIN_FILE_SUFFIX)
-		if os.path.isfile(file_path):
-			os.rename(file_path, new_file_path)
-			self.load_plugin(new_file_path)
+		with self.mani_lock:
+			self.logger.info(self.mcdr_server.tr('plugin_manager.enable_plugin.entered', file_path))
+			new_file_path = string_util.remove_suffix(file_path, constant.DISABLED_PLUGIN_FILE_SUFFIX)
+			if os.path.isfile(file_path):
+				os.rename(file_path, new_file_path)
+				self.load_plugin(new_file_path)
 
 	def disable_plugin(self, plugin: RegularPlugin):
-		self.logger.info(self.mcdr_server.tr('plugin_manager.disable_plugin.entered', plugin))
-		self.unload_plugin(plugin)
-		if os.path.isfile(plugin.file_path):
-			os.rename(plugin.file_path, plugin.file_path + constant.DISABLED_PLUGIN_FILE_SUFFIX)
+		with self.mani_lock:
+			self.logger.info(self.mcdr_server.tr('plugin_manager.disable_plugin.entered', plugin))
+			self.unload_plugin(plugin)
+			if os.path.isfile(plugin.file_path):
+				os.rename(plugin.file_path, plugin.file_path + constant.DISABLED_PLUGIN_FILE_SUFFIX)
 
 	def refresh_all_plugins(self):
-		self.logger.info(self.mcdr_server.tr('plugin_manager.refresh_all_plugins.entered'))
-		return self.__refresh_plugins(lambda plg: True)
+		with self.mani_lock:
+			self.logger.info(self.mcdr_server.tr('plugin_manager.refresh_all_plugins.entered'))
+			return self.__refresh_plugins(lambda plg: True)
 
 	def refresh_changed_plugins(self):
-		self.logger.info(self.mcdr_server.tr('plugin_manager.refresh_changed_plugins.entered'))
-		return self.__refresh_plugins(lambda plg: plg.file_changed())
+		with self.mani_lock:
+			self.logger.info(self.mcdr_server.tr('plugin_manager.refresh_changed_plugins.entered'))
+			return self.__refresh_plugins(lambda plg: plg.file_changed())
 
 	# ----------------
 	#   Plugin Event
 	# ----------------
 
-	def dispatch_event(self, event: MCDREvent, args: Tuple[Any, ...]):
+	def __dispatch_event(self, event: MCDREvent, args: Tuple[Any, ...]):
 		self.logger.debug('Dispatching {} with args ({})'.format(event, ', '.join([type(arg).__name__ for arg in args])), option=DebugOption.PLUGIN)
 		for listener in self.registry_storage.event_listeners.get(event.id, []):
 			self.trigger_listener(listener, args)
 
+	def dispatch_event(self, event: MCDREvent, args: Tuple[Any, ...], *, on_executor_thread=True, wait=False):
+		if on_executor_thread:
+			self.mcdr_server.task_executor.execute_or_enqueue(lambda: self.__dispatch_event(event, args), wait=wait)
+		else:  # on thread
+			self.__dispatch_event(event, args)
+
 	def trigger_listener(self, listener: EventListener, args: Tuple[Any, ...]):
 		"""
 		The terminated entry for triggering a listener
 		The server_interface parameter will be automatically added as the 1st parameter
 		"""
 		# self.thread_pool.add_task(lambda: listener.execute(*args), listener.plugin)
-		self.set_current_plugin(listener.plugin)
-		try:
-			listener.execute(self.mcdr_server.server_interface, *args)
-		except:
-			self.logger.exception('Error invoking listener {}'.format(listener))
-		finally:
-			self.set_current_plugin(None)
+		with self.with_plugin_context(listener.plugin):
+			try:
+				listener.execute(self.mcdr_server.server_interface, *args)
+			except:
+				self.logger.exception('Error invoking listener {}'.format(listener))
